"""
issue action

Creates a new github issue and asana task.

"""

import logging

from ..action import Action

class Issue(Action):
    """Creates a new github issue and asana task."""

    # name of action
    name = "issue"

    @classmethod
    def add_arguments(cls, parser):
        """Add arguments to the parser for collection in app.args.

        Args:
            parser:
                `argparse.ArgumentParser`. Parser.
                Arguments added here are server on
                self.args.
        """

        parser.add_argument(
            '-t', '--title',
            action='store',
            nargs='?',
            const='',
            dest='title',
            help="[issue] task/issue title.",
            )

        parser.add_argument(
            '-b', '--body',
            action='store',
            nargs='?',
            const='',
            dest='body',
            help="[issue] task/issue body.",
            )

        pass

    def run(self):
        app = self.app

        # OAuth 2 exchange.
        app.authenticate()

        repo, project = self.get_repo_and_project()

        # Collect title and body
        title = app.settings.apply(None, app.args.title,
            "task/issue title",
            )

        assert title, "title required"

        body = app.settings.apply(None, app.args.body,
            "body/message",
            ) or ''

        # Post asana task.
        asana_workspace_id = project['workspace']['id']
        task = app.asana.tasks.create_in_workspace(
            asana_workspace_id,
            {
            'name': title,
            'notes': body,
            # TODO: Correct assignee.
            'assignee': 'me',
            'projects': [project['id']]
            })

        asana_task_id = task['id']
        asana_task_url = app.make_asana_url(project['id'], asana_task_id)

        body = body + ("\n\n"
            "**Asana: #%d**\n"
            "%s" % (
                asana_task_id,
                asana_task_url,
            ))

        # Create github issue
        issue = repo.create_issue(
            title=title,
            body=body.strip(),
            )

        # Create asana comment (story)
        app.announce_issue_to_task(asana_task_id, issue)

        logging.info("github issue #%d created:\n%s\n",
            issue.number, issue.html_url)

        logging.info("asana task #%d created:\n%s\n",
            asana_task_id, asana_task_url)

        # Save issue and task to db.
        app.save_issue_data_task(issue.number, asana_task_id)



"""
Command-line interface for asana-hub.
"""

import argparse
import logging
import sys
import os
import traceback

try:
    from asana import Client
    from asana import error as asana_errors
    from github import Github

    import urllib3
    import certifi

    # Setup pool manager and SSL verification.
    http = urllib3.PoolManager(
        cert_reqs='CERT_REQUIRED', # Force certificate check.
        ca_certs=certifi.where(),  # Path to the Certifi bundle.
    )

    try:
        import urllib3.contrib.pyopenssl
        urllib3.contrib.pyopenssl.inject_into_urllib3()
    except:
        logging.debug("pyopenssl not detected.\n"
            "to install pyopenssl: "
            "pip install pyopenssl ndg-httpsclient pyasn1")

except ImportError:
    raise Exception("Could not import required packages.\n"
        "Did you pip install -r requirements.txt ?")

from .json_data import JSONData
from .action import Action

class ToolApp(object):

    def authenticate(self):
        """Connects to Github and Asana and authenticates via OAuth."""
        if self.oauth:
            return False

        # Save asana.
        self.settings.apply('api-asana', self.args.asana_api,
            "enter asana api key")

        # Save github.com
        self.settings.apply('api-github', self.args.github_api,
            "enter github.com token")

        logging.debug("authenticating asana api.")
        self.asana = Client.basic_auth(self.settings['api-asana'])
        self.asana_errors = asana_errors
        self.asana_me = self.asana.users.me()
        logging.debug("authenticating github api")
        self.github = Github(self.settings['api-github'])
        self.github_user = self.github.get_user()

        self.oauth = True

    @classmethod
    def _list_select(cls, lst, prompt, offset=0):
        """Given a list of values and names, accepts the index value or name."""

        inp = raw_input("select %s: " % prompt)
        assert inp, "value required."

        try:
            return lst[int(inp)+offset]
        except ValueError:
            return inp
        except IndexError:
            assert False, "bad value."

    def prompt_repo(self):

        # Select org
        orgs = [None]
        print "0) %s [you]" % self.github_user.name
        for idx, org in enumerate(self.github_user.get_orgs()):
            orgs.append(org)
            print "%d) %s" % (idx+1, org.name)

        org = self._list_select(orgs, "org")
        ctx = org or self.github_user

        repos = []
        for idx, repo in enumerate(ctx.get_repos()):
            repos.append(repo)
            print "%d) %s" % (idx, repo.name)

        repo = self._list_select(repos, "repo")
        logging.debug("repo id: %d", repo.id)

        return repo

    def prompt_project(self):

        # Select workspace
        workspaces = []
        for idx, workspace in enumerate(self.asana_me['workspaces']):
            workspaces.append(workspace)
            print "%d) %s" % (idx, workspace['name'])

        workspace = self._list_select(workspaces, "workspace")

        # Select workspace
        as_projects = self.asana.projects.find_by_workspace(workspace['id'],
            iterator_type=None)

        projects = []

        for idx, project in enumerate(as_projects):
            projects.append(project)
            print "%d) %s" % (idx, project['name'])

        project = self._list_select(projects, "project")

        return project

    @classmethod
    def make_asana_url(cls, project_id, task_id):
        """Returns a URL to an asana task."""
        return "https://app.asana.com/0/%d/%d" % (project_id, task_id)

    ##################
    ### Issue Data ###
    ##################

    @classmethod
    def _issue_data_key(cls, namespace):
        """Returns key for issue_data in data."""
        return 'issue_data_%s' % namespace

    def save_issue_data_task(self, issue, task, namespace='open'):
        """Saves a issue data (tasks, etc.) to local data.

        Args:
            issue:
                `int`. Github issue number.
            task:
                `int`. Asana task ID.
            namespace:
                `str`. Namespace for storing this issue.
        """

        issue_data = self.get_saved_issue_data(issue, namespace)

        if not issue_data.has_key('tasks'):
            issue_data['tasks'] = [task]
        elif task not in issue_data['tasks']:
            issue_data['tasks'].append(task)

    def has_saved_issue_data(self, issue, namespace='open'):
        issue_data_key = self._issue_data_key(namespace)
        issue_data = self.data.get(issue_data_key,
            {})

        if isinstance(issue, int):
            issue_number = str(issue)
        elif isinstance(issue, basestring):
            issue_number = issue
        else:
            issue_number = issue.number

        return issue_data.has_key(str(issue_number))

    def get_saved_issue_data(self, issue, namespace='open'):
        """Returns issue data from local data.

        Args:
            issue:
                `int`. Github issue number.
            namespace:
                `str`. Namespace for storing this issue.
        """

        if isinstance(issue, int):
            issue_number = str(issue)
        elif isinstance(issue, basestring):
            issue_number = issue
        else:
            issue_number = issue.number

        issue_data_key = self._issue_data_key(namespace)
        issue_data = self.data.get(issue_data_key,
            {})

        _data = issue_data.get(str(issue_number), {})
        issue_data[str(issue_number)] = _data
        return _data

    def move_saved_issue_data(self, issue, ns, other_ns):
        """Moves an issue_data from one namespace to another."""

        if isinstance(issue, int):
            issue_number = str(issue)
        elif isinstance(issue, basestring):
            issue_number = issue
        else:
            issue_number = issue.number

        issue_data_key = self._issue_data_key(ns)
        other_issue_data_key = self._issue_data_key(other_ns)
        issue_data = self.data.get(issue_data_key,
            {})
        other_issue_data = self.data.get(other_issue_data_key,
            {})

        _id = issue_data.pop(issue_number, None)
        if _id:
            other_issue_data[issue_number] = _id

        self.data[other_issue_data_key] = other_issue_data
        self.data[issue_data_key] = issue_data

    #################
    ### Task Data ###
    #################

    @classmethod
    def _task_data_key(cls):
        """Returns key for task_data in data."""
        return 'task-data'

    def has_saved_task_data(self, task):
        task_data_key = self._task_data_key()
        task_data = self.data.get(task_data_key,
            {})

        if isinstance(task, int):
            task_number = str(task)
        elif isinstance(task, basestring):
            task_number = task
        else:
            task_number = task['id']

        return task_data.has_key(str(task_number))

    def get_saved_task_data(self, task):
        """Returns task data from local data.

        Args:
            task:
                `int`. Asana task number.
        """

        if isinstance(task, int):
            task_number = str(task)
        elif isinstance(task, basestring):
            task_number = task
        else:
            task_number = task['id']

        task_data_key = self._task_data_key()
        task_data = self.data.get(task_data_key, {})

        _data = task_data.get(str(task_number), {})
        task_data[str(task_number)] = _data
        return _data

    #############
    ### Misc. ###
    #############

    def announce_issue_to_task(self, asana_task_id, issue):
        """Creates a story on a task announcing the issue."""
        return self.asana.stories.create_on_task(asana_task_id,
            {
            'text':
                "Git Issue #%d: \n"
                "%s" % (
                    issue.number,
                    issue.html_url,
                    )
            })

    def get_asana_task(self, asana_task_id):
        """Retrieves a task from asana."""

        try:
            return self.asana.tasks.find_by_id(asana_task_id)
        except asana_errors.NotFoundError:
            return None

    def __init__(self, version):
        """Accepts version of the app."""

        # Setup settings
        self.version = version
        self.exit_code = 999
        self.oauth = False

        # Setup logging
        self.logger = logging.getLogger()
        formatter = logging.Formatter("%(message)s")
        self.logger.setLevel(logging.DEBUG)
        self.logger.propagate = False
        self.logger.handlers = []
        ch = logging.StreamHandler()
        ch.setLevel(logging.INFO)
        ch.setFormatter(formatter)
        self.logger.addHandler(ch)

        parser = argparse.ArgumentParser(description='Loudr Utility Tool')

        # Default paths for settings and data
        def_settings_file = os.path.join(
            os.path.expanduser("~"),
            '.asana-hub',
            )

        def_data_file = os.path.expanduser("./.asana-hub.proj")

        # Load actions
        actions = {}
        choices = []
        help_msgs = ''

        for action in Action.iter_actions():
            actions[action.name] = action
            choices.append(action.name)
            help_msgs += '\n-  %s: %s' % (action.name, action.__doc__)

        help_msgs += '\n'

        parser.add_argument(
            'action',
            action='store',
            nargs=1,
            help=help_msgs,
            choices=choices,
            )

        parser.add_argument(
            '-vv', '--verbose',
            action='store_true',
            dest='verbose',
            help="use debugging verbosity.",
            )

        parser.add_argument(
            '-s', '--settings-file',
            action='store',
            nargs='?',
            dest='settings_file',
            default=def_settings_file,
            help="path to save oauth api keys.",
            )

        parser.add_argument(
            '-d', '--data-file',
            action='store',
            nargs='?',
            dest='data_file',
            default=def_data_file,
            help="path to save repository and project based data.",
            )

        # Add action arguments.
        for action in actions.values():
            action.add_arguments(parser)

        # Add actions from the parent class. (The settings)
        Action.add_arguments(parser)

        parser.add_argument('-v', '--version', action='version',
            version='%(prog)s ' + '%s' % version)

        self.args = parser.parse_args()

        if len(sys.argv) < 2:
            parser.print_help()
            self.exit_code = 1
            return

        if self.args.verbose:
            ch.setLevel(logging.DEBUG)

        # Load settings
        self.settings = JSONData(filename=self.args.settings_file,
            args=self.args, version=version)

        self.data = JSONData(filename=self.args.data_file,
            args=self.args, version=version)

        # Load action method and call.
        try:
            action_name = self.args.action[0]
            action_class = actions.get(action_name)
            if not action_class:
                raise NotImplementedError(
                    "%s is not implemented." % action_name)

            # Instantiate and run
            action = action_class(app=self, args=self.args)
            action.run()
        except AssertionError as exc:
            logging.error("Error: %s", unicode(exc))
            logging.debug("%s", traceback.format_exc())
            self.exit_code = 1
            return
        except Exception as exc:
            logging.exception("Exception: %r", exc)
            self.exit_code = 129
            return
        finally:

            # Save settings
            self.settings.save()
            # Save data
            self.data.save()

        self.exit_code = 0




# -*- coding: utf-8 -*-
"""
sync action

Syncs completion status of issues and their matched tasks.

"""

import logging
import re
import collections

from ..action import Action

ASANA_ID_RE = re.compile(r'#(\d{12,16})', re.M)
"""Regular expression for capturing asana IDs."""

ASANA_SECTION_RE = re.compile(r'## Asana Tasks:\s+(#(\d{12,})\s*)+', re.M)
"""Regular exprsssion to catch malformed data due to too many tasks."""

_ms_label = lambda x: "_ms:%d"%x
"""Converts a milestone id into an _ms prefixed string"""

class Sync(Action):
    """Syncs completion status of issues and their matched tasks."""

    # name of action
    name = "sync"

    @classmethod
    def add_arguments(cls, parser):
        """Add arguments to the parser for collection in app.args.

        Args:
            parser:
                `argparse.ArgumentParser`. Parser.
                Arguments added here are server on
                self.args.
        """

        parser.add_argument(
            '-c', '--create-missing-tasks',
            action='store_true',
            dest='create_missing_tasks',
            help="[sync] create asana tasks for issues without tasks"
            )

        parser.add_argument(
            '-l', '--sync-labels',
            action='store_true',
            dest='sync_labels',
            help="[sync] sync labels and milestones for each issue"
            )

    def apply_tasks_to_issue(self, issue, tasks, issue_body=None):
        """Applies task numbers to an issue."""
        issue_body = issue_body or issue.body
        task_numbers = "\n".join('#'+str(tid) for tid in tasks)
        if task_numbers:
            new_body = ASANA_SECTION_RE.sub('', issue_body)
            new_body = new_body + "\n## Asana Tasks:\n\n%s" % task_numbers
            issue.edit(body=new_body)
            return new_body

        return issue_body

    def sync_labels(self, repo):
        """Creates a local map of github labels/milestones to asana tags."""

        logging.info("syncing new github.com labels to tags")

        # create label tag map
        ltm = self.app.data.get("label-tag-map", {})

        # loop over labels, if they don't have tags, make them
        for label in repo.get_labels():
            tag_id = ltm.get(label.name, None)
            if tag_id is None:

                tag = self.app.asana.tags.create(name=label.name,
                                      workspace=self.asana_ws_id,
                                      notes="gh: %s" % label.url
                                      )

                logging.info("\t%s => tag %d", label.name, tag['id'])
                ltm[label.name] = tag['id']

        # loop over milestones, if they don't have tags, make them
        for ms in repo.get_milestones(state="all"):
            tag_id = ltm.get(_ms_label(ms.id), None)
            if tag_id is None:

                tag = self.app.asana.tags.create(name=ms.title,
                                      workspace=self.asana_ws_id,
                                      notes="gh: %s" % ms.url
                                      )

                logging.info("\t%s => tag %d", ms.title, tag['id'])
                ltm[_ms_label(ms.id)] = tag['id']

        self.app.data['label-tag-map'] = ltm
        return ltm

    def run(self):
        app = self.app

        # OAuth 2 exchange.
        app.authenticate()

        repo, project = self.get_repo_and_project()
        self.asana_ws_id = asana_workspace_id = project['workspace']['id']
        project_id = project['id']

        # Sync project labels <-> asana tags
        if app.args.sync_labels:
            label_tag_map = self.sync_labels(repo)
        else:
            label_tag_map = {}

        # Iterate over the issues in the opposite state as the namespace
        # we are in. We simply want to toggle these guys.
        logging.info("collecting github.com issues")
        issues_map = {}

        # Get the first issue, to limit syncing.
        first_issue = app.data.get('first-issue')

        for issue in repo.get_issues(state="all"):

            # bypass issues < `first-issue` setting.
            if (first_issue is not None and
                issue.number < first_issue):
                logging.debug("stopping at first-issue: %d", first_issue)
                break

            issue_number = str(issue.number)
            issue_body = issue.body
            asana_match = ASANA_ID_RE.search(issue_body)
            multi_match_sections = len(ASANA_SECTION_RE.findall(issue_body)) > 1

            state = issue.state
            other_state = 'open' if state == 'closed' else 'closed'
            status = "cached"

            # Collect closed and opened tasks known for this issue.
            closed_tasks = \
                app.get_saved_issue_data(issue, 'closed').get('tasks', [])
            open_tasks = \
                app.get_saved_issue_data(issue, 'open').get('tasks', [])

            recorded_tasks = set(open_tasks + closed_tasks)

            # Collect tasks named on github issue
            issue_named_tasks = set()
            for m_grp in ASANA_ID_RE.finditer(issue_body):
                issue_named_tasks.add(int(m_grp.group(1)))

            # Get tasks that are named but missing from cache.
            tasks_to_save_to_this_issue = issue_named_tasks - recorded_tasks
            for task_id in tasks_to_save_to_this_issue:
                status = "collected tasks"
                app.save_issue_data_task(issue_number, task_id, issue.state)

            my_tasks = recorded_tasks.union(tasks_to_save_to_this_issue)

            # Determine if there are multiple groups of ASANA TASKS
            # named.
            if multi_match_sections:
                issue_body = ASANA_SECTION_RE.sub('', issue_body)
                asana_match = None

                issue_body = self.apply_tasks_to_issue(issue, my_tasks,
                    issue_body=issue_body)
                status = "minified issue body"

            # Sync tags and labels
            labels = set()
            if app.args.sync_labels:
                for label in issue.get_labels():
                    labels.add(label.name)
                if issue.milestone:
                    labels.add(_ms_label(issue.milestone.id))

            # If we have tasks already, this issue is cached.
            if recorded_tasks:
                issue_data = app.get_saved_issue_data(issue_number, other_state)
                issue_cached_tasks = issue_data.get('tasks', [])

                issues_map[issue_number] = (issue, issue_cached_tasks)

                # If the body is missing asana tasks, add all those we know
                # about.
                if not asana_match:
                    # Add tasks if we have any.
                    if recorded_tasks:
                        issue_body = self.apply_tasks_to_issue(issue, my_tasks,
                            issue_body=issue_body)
                        status = "updated with asana #s"

                # If the section isn't formatted... let's reformat it.
                elif not ASANA_SECTION_RE.search(issue_body):
                    issue_body = self.apply_tasks_to_issue(issue, my_tasks,
                        issue_body=issue_body)
                    status = "reformatted asana tasks"


            # tasks named on issue need to be synced
            elif asana_match and issue_named_tasks:
                status = "connecting tasks"
                self.apply_tasks_to_issue(issue, my_tasks,
                    issue_body=issue_body)

                issues_map[issue_number] = (issue, my_tasks)

            elif self.args.create_missing_tasks and not issue.pull_request:
                # missing task
                # Create tasks for non-prs
                task = app.asana.tasks.create_in_workspace(
                    asana_workspace_id,
                    {
                        'name': issue.title,
                        'notes': issue_body,
                        # TODO: Correct assignee.
                        'assignee': 'me',
                        'projects': [project_id],
                        'completed': bool(issue.closed_at)
                    })

                # Announce task git issue
                task_id = task['id']
                app.announce_issue_to_task(task_id, issue)

                # Save task to drive
                app.save_issue_data_task(issue_number, task_id,
                    issue.state)
                status = "new task #%d" % task_id

                my_tasks.add(task_id)

            else:
                status = "no task"

            logging.info("\t%d) %s - %s",
                issue.number, issue.title, status)

            # Sync tags/labels
            for task in my_tasks:
                task_data = app.get_saved_task_data(task)
                tag_ids = task_data.get('tags') or []
                try:
                    added_tags = 0
                    for label in labels:
                        tag_id = label_tag_map.get(label)
                        if not tag_id:
                            continue
                        if tag_id in tag_ids:
                            continue
                        tag_ids.append(tag_id)
                        app.asana.tasks.add_tag(task, tag=tag_id)
                        added_tags += 1

                    if added_tags:
                        logging.info("\t\t - added %d tags to %s",
                                     added_tags, task)
                        task_data['tags'] = tag_ids

                except app.asana_errors.InvalidRequestError:
                    logging.warn("warning: bad task %d", task)

        # Refresh status of issue/tasks
        logging.info("refreshing task statuses...")
        for issue_number, (issue, issue_tasks) in issues_map.iteritems():

            state = issue.state
            other_state = 'open' if state == 'closed' else 'closed'

            # Get tasks in the issue that are outdated.
            for task_id in issue_tasks:
                logging.info("\t#%d - updating (%s->%s) - %d",
                    issue.number,
                    other_state, state,
                    task_id)

                task = app.get_asana_task(task_id)
                if not task:
                    issue_tasks.remove(task_id)
                    logging.debug("task #%d was deleted.", task_id)
                    continue

                if issue.closed_at:
                    # close task
                    app.asana.tasks.update(
                        task['id'],
                        {
                        'completed': True
                        })
                else:
                    # open task
                    app.asana.tasks.update(
                        task['id'],
                        {
                        'completed': False
                        })

                app.move_saved_issue_data(issue_number, other_state, state)







"""
issue action

Creates a new github issue and asana task.

"""

import logging

from ..action import Action

class PullRequest(Action):
    """Creates a new github pull-request for an exist issue and a branch."""

    # name of action
    name = "pr"

    @classmethod
    def add_arguments(cls, parser):
        """Add arguments to the parser for collection in app.args.

        Args:
            parser:
                `argparse.ArgumentParser`. Parser.
                Arguments added here are server on
                self.args.
        """

        parser.add_argument(
            '-i', '--issue',
            action='store',
            nargs='?',
            const='',
            dest='issue',
            help="[pr] issue #",
            )

        parser.add_argument(
            '-br', '--branch',
            action='store',
            nargs='?',
            const='',
            dest='branch',
            help="[pr] branch",
            )

        parser.add_argument(
            '-tbr', '--target-branch',
            action='store',
            nargs='?',
            const='',
            default='master',
            dest='target_branch',
            help="[pr] name of branch to pull changes into\n(defaults to: master)",
            )

    def run(self):
        app = self.app

        # OAuth 2 exchange.
        app.authenticate()

        repo, project = self.get_repo_and_project()
        project_id = project['id']

        # Collect title and body
        issue = app.settings.apply(None, app.args.issue,
            "issue # to create PR for",
            )

        assert issue, "issue required"
        issue = repo.get_issue(int(issue))
        assert issue, "issue could not be found"

        branch = app.settings.apply(None, app.args.branch,
            "branch to create pull-request from [via api]",
            ) or ''

        # Get issue data to create pull request tasks list.
        issue_data = app.get_saved_issue_data(issue)

        issue_tasks = issue_data.get('tasks', [])
        issue_data['tasks'] = issue_tasks

        # pull_requests is a list of pull request numbers
        issue_prs = issue_data.get('pull_requests', [])
        issue_data['pull_requests'] = issue_prs

        asana_msgs = ''
        for task in issue_tasks:
            asana_msgs += '\n * [#%d](%s)' % (
                task,
                app.make_asana_url(project_id, task)
                )

        if asana_msgs:
            asana_msgs = '## Asana Tasks\n' + asana_msgs

        # Create pull request.
        pull_request = repo.create_pull(
            title=issue.title,
            head=branch,
            base="master",
            body="Fixes #%d - %s\n"
                "\n## Testing:\n\n"
                "%s"
                "" % (
                    issue.number,
                    issue.title,
                    asana_msgs,
                    )
            )

        # Post asana task.
        title = "PR #%d" % pull_request.number
        body = "%s" % (
            pull_request.html_url
            )

        task_ids = []
        for task_id in issue_tasks:
            task = app.asana.tasks.add_subtask(
                task_id,
                {
                'name': title,
                'notes': body,
                # TODO: Correct assignee.
                'assignee': 'me',
                })
            task_ids.append(task['id'])

            app.save_issue_data_task(issue.number, task['id'])

        # Add pull request to local data
        if pull_request.number not in issue_prs:
            issue_prs.append(pull_request.number)

        logging.info("github pull_request #%d created:\n%s\n",
            pull_request.number, pull_request.html_url)

